"""Tests for changelog generator."""

import gc
import time
from datetime import datetime
from pathlib import Path

import git
import pytest

from changelog_maestro.core.config import Config
from changelog_maestro.core.generator import ChangelogGenerator
from changelog_maestro.core.git import GitRepository
from changelog_maestro.utils.exceptions import ChangelogError


class TestChangelogGenerator:
    """Test cases for ChangelogGenerator."""

    def test_init(self, temp_repo: Path):
        """Test initializing generator."""
        config = Config()
        generator = ChangelogGenerator(temp_repo, config)

        assert generator.repo_path == temp_repo
        assert generator.config == config
        assert isinstance(generator.git_repo, GitRepository)

    def test_generate_with_conventional_commits(self, temp_repo: Path):
        """Test generating changelog with conventional commits."""
        # Add conventional commits to repository
        repo = git.Repo(temp_repo)

        try:
            # Add feature commit
            feature_file = temp_repo / "feature.txt"
            feature_file.write_text("new feature")
            repo.index.add([str(feature_file)])
            repo.index.commit("feat: add new feature")

            # Add bug fix commit
            bug_file = temp_repo / "bug.txt"
            bug_file.write_text("bug fix")
            repo.index.add([str(bug_file)])
            repo.index.commit("fix: resolve critical bug")

            # Generate changelog
            generator = ChangelogGenerator(temp_repo)
            result = generator.generate()

            assert "# Changelog" in result
            assert "Unreleased" in result
            assert "Features" in result
            assert "Bug Fixes" in result
            assert "add new feature" in result
            assert "resolve critical bug" in result
        finally:
            repo.close()
            gc.collect()
            time.sleep(0.1)

    def test_generate_with_tags(self, temp_repo: Path):
        """Test generating changelog with version tags."""
        repo = git.Repo(temp_repo)

        try:
            # Add feature and create tag
            feature_file = temp_repo / "feature.txt"
            feature_file.write_text("new feature")
            repo.index.add([str(feature_file)])
            repo.index.commit("feat: add new feature")
            repo.create_tag("v1.0.0")

            # Add bug fix
            bug_file = temp_repo / "bug.txt"
            bug_file.write_text("bug fix")
            repo.index.add([str(bug_file)])
            repo.index.commit("fix: resolve bug")

            # Generate changelog
            generator = ChangelogGenerator(temp_repo)
            result = generator.generate()

            assert "## [1.0.0]" in result
            assert "add new feature" in result
            assert "resolve bug" in result
        finally:
            repo.close()
            gc.collect()
            time.sleep(0.1)

    def test_generate_no_commits(self, temp_repo: Path):
        """Test generating changelog with no valid commits."""
        repo = git.Repo(temp_repo)

        try:
            # Add non-conventional commit
            test_file = temp_repo / "test.txt"
            test_file.write_text("test")
            repo.index.add([str(test_file)])
            repo.index.commit("not a conventional commit")

            generator = ChangelogGenerator(temp_repo)

            with pytest.raises(
                ChangelogError, match="No valid conventional commits found"
            ):
                generator.generate()
        finally:
            repo.close()
            gc.collect()
            time.sleep(0.1)

    def test_generate_with_breaking_changes(self, temp_repo: Path):
        """Test generating changelog with breaking changes."""
        repo = git.Repo(temp_repo)

        try:
            # Add breaking change commit
            breaking_file = temp_repo / "breaking.txt"
            breaking_file.write_text("breaking change")
            repo.index.add([str(breaking_file)])
            repo.index.commit("feat!: breaking change in API")

            generator = ChangelogGenerator(temp_repo)
            result = generator.generate()

            assert "⚠ BREAKING CHANGES" in result
            assert "breaking change in API" in result
        finally:
            repo.close()
            gc.collect()
            time.sleep(0.1)

    def test_generate_with_exclude_patterns(self, temp_repo: Path):
        """Test generating changelog with exclude patterns."""
        repo = git.Repo(temp_repo)

        try:
            # Add commits using proper file operations
            feature_file = temp_repo / "feature.txt"
            feature_file.write_text("feature")
            repo.index.add([str(feature_file)])
            repo.index.commit("feat: add feature")

            bug_file = temp_repo / "bug.txt"
            bug_file.write_text("bug fix")
            repo.index.add([str(bug_file)])
            repo.index.commit("fix: resolve bug")

            chore_file = temp_repo / "chore.txt"
            chore_file.write_text("chore")
            repo.index.add([str(chore_file)])
            repo.index.commit("chore: update dependencies")

            # Configure to exclude chore commits
            config = Config(exclude_patterns=["chore"])
            generator = ChangelogGenerator(temp_repo, config)
            result = generator.generate()

            assert "add feature" in result
            assert "resolve bug" in result
            assert "update dependencies" not in result
        finally:
            repo.close()
            gc.collect()
            time.sleep(0.1)

    def test_generate_with_custom_sections(self, temp_repo: Path):
        """Test generating changelog with custom sections."""
        repo = git.Repo(temp_repo)

        try:
            # Add commits of different types
            feature_file = temp_repo / "feature.txt"
            feature_file.write_text("feature")
            repo.index.add([str(feature_file)])
            repo.index.commit("feat: add feature")

            bug_file = temp_repo / "bug.txt"
            bug_file.write_text("bug fix")
            repo.index.add([str(bug_file)])
            repo.index.commit("fix: resolve bug")

            docs_file = temp_repo / "docs.txt"
            docs_file.write_text("docs")
            repo.index.add([str(docs_file)])
            repo.index.commit("docs: update documentation")

            # Configure to only include features and fixes
            config = Config(sections=["feat", "fix"])
            generator = ChangelogGenerator(temp_repo, config)
            result = generator.generate()

            assert "add feature" in result
            assert "resolve bug" in result
            assert "update documentation" not in result
        finally:
            repo.close()
            gc.collect()
            time.sleep(0.1)

    def test_generate_json_format(self, temp_repo: Path):
        """Test generating changelog in JSON format."""
        repo = git.Repo(temp_repo)

        try:
            # Add conventional commit
            feature_file = temp_repo / "feature.txt"
            feature_file.write_text("feature")
            repo.index.add([str(feature_file)])
            repo.index.commit("feat: add new feature")

            # Configure JSON output
            config = Config(output_style="json")
            generator = ChangelogGenerator(temp_repo, config)
            result = generator.generate()

            assert '"changelog"' in result
            assert '"type": "feat"' in result
            assert '"description": "add new feature"' in result
        finally:
            repo.close()
            gc.collect()
            time.sleep(0.1)

    def test_generate_yaml_format(self, temp_repo: Path):
        """Test generating changelog in YAML format."""
        repo = git.Repo(temp_repo)

        try:
            # Add conventional commit
            feature_file = temp_repo / "feature.txt"
            feature_file.write_text("feature")
            repo.index.add([str(feature_file)])
            repo.index.commit("feat: add new feature")

            # Configure YAML output
            config = Config(output_style="yaml")
            generator = ChangelogGenerator(temp_repo, config)
            result = generator.generate()

            assert "changelog:" in result
            assert "type: feat" in result
            assert "description: add new feature" in result
        finally:
            repo.close()
            gc.collect()
            time.sleep(0.1)



"""Pytest configuration and fixtures."""

import gc
import os
import shutil
import tempfile
import time
from datetime import datetime
from pathlib import Path
from typing import Generator

import git
import pytest

from changelog_maestro.core.config import Config
from changelog_maestro.core.git import Commit, GitRepository


def force_remove_readonly(func, path, exc):
    """Force remove readonly files on Windows."""
    if os.path.exists(path):
        os.chmod(path, 0o777)
        func(path)


@pytest.fixture
def temp_repo() -> Generator[Path, None, None]:
    """Create a temporary Git repository for testing."""
    temp_dir = tempfile.mkdtemp()
    repo_path = Path(temp_dir)
    repo = None

    try:
        # Initialize Git repository
        repo = git.Repo.init(repo_path)

        # Configure Git user
        repo.config_writer().set_value("user", "name", "Test User").release()
        repo.config_writer().set_value("user", "email", "test@example.com").release()

        # Create initial commit
        readme_file = repo_path / "README.md"
        readme_file.write_text("# Test Repository\n")
        repo.index.add([str(readme_file)])
        repo.index.commit("Initial commit")

        yield repo_path

    finally:
        # Aggressive cleanup for Windows
        try:
            if repo is not None:
                repo.close()
                del repo
            gc.collect()
            time.sleep(0.2)  # Give Windows more time

            # Force remove the directory
            if os.path.exists(temp_dir):
                shutil.rmtree(temp_dir, onerror=force_remove_readonly)
        except Exception:
            # If cleanup fails, try again after a longer delay
            try:
                time.sleep(0.5)
                if os.path.exists(temp_dir):
                    shutil.rmtree(temp_dir, onerror=force_remove_readonly)
            except Exception:
                pass  # Give up if we still can't clean up


@pytest.fixture
def sample_commits() -> list[Commit]:
    """Sample commits for testing."""
    return [
        Commit(
            hash="abc123",
            message="feat: add new feature",
            author="Test User",
            email="test@example.com",
            date=datetime(2023, 1, 1, 12, 0, 0),
        ),
        Commit(
            hash="def456",
            message="fix: resolve bug in parser",
            author="Test User",
            email="test@example.com",
            date=datetime(2023, 1, 2, 12, 0, 0),
        ),
        Commit(
            hash="ghi789",
            message="feat!: breaking change in API",
            author="Test User",
            email="test@example.com",
            date=datetime(2023, 1, 3, 12, 0, 0),
        ),
    ]


@pytest.fixture
def default_config() -> Config:
    """Default configuration for testing."""
    return Config()


@pytest.fixture
def git_repo(temp_repo: Path) -> GitRepository:
    """GitRepository instance for testing."""
    return GitRepository(temp_repo)



"""Conventional Commits parser."""

import re
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from ..utils.exceptions import ParseError


@dataclass
class ParsedCommit:
    """Represents a parsed conventional commit."""

    type: str
    scope: Optional[str]
    description: str
    body: Optional[str]
    footer: Optional[str]
    is_breaking: bool
    breaking_change_description: Optional[str] = None

    @property
    def is_feature(self) -> bool:
        """Check if commit is a feature."""
        return self.type == "feat"

    @property
    def is_fix(self) -> bool:
        """Check if commit is a bug fix."""
        return self.type == "fix"

    @property
    def formatted_scope(self) -> str:
        """Get formatted scope for display."""
        return f"({self.scope})" if self.scope else ""

    @property
    def short_description(self) -> str:
        """Get short description (first line of description)."""
        return self.description.split("\n")[0] if self.description else ""


class ConventionalCommitParser:
    """Parser for Conventional Commits specification."""

    # Regex pattern for conventional commit format
    COMMIT_PATTERN = re.compile(
        r"^(?P<type>\w+)"
        r"(?:\((?P<scope>[^)]+)\))?"
        r"(?P<breaking>!)?"
        r": (?P<description>.+?)(?:\n\n(?P<body>.*))?$",
        re.DOTALL,
    )

    # Breaking change footer patterns
    BREAKING_CHANGE_PATTERNS = [
        re.compile(r"^BREAKING CHANGE:\s*(.+)", re.MULTILINE | re.IGNORECASE),
        re.compile(r"^BREAKING-CHANGE:\s*(.+)", re.MULTILINE | re.IGNORECASE),
    ]

    def __init__(self, breaking_change_indicators: Optional[List[str]] = None) -> None:
        """Initialize parser with optional custom breaking change indicators."""
        self.breaking_change_indicators = breaking_change_indicators or [
            "BREAKING CHANGE",
            "BREAKING-CHANGE",
        ]

    def parse(self, commit_message: str) -> ParsedCommit:
        """Parse a commit message according to Conventional Commits spec."""
        if not commit_message.strip():
            raise ParseError("Empty commit message")

        # Clean up the commit message
        message = commit_message.strip()

        # Try to match the conventional commit pattern
        match = self.COMMIT_PATTERN.match(message)
        if not match:
            raise ParseError(
                f"Commit message does not follow Conventional Commits format: {message[:50]}..."
            )

        groups = match.groupdict()

        # Extract basic components
        commit_type = groups["type"].lower()
        scope = groups.get("scope")
        description = groups["description"].strip()
        body = groups.get("body", "").strip() if groups.get("body") else None

        # Check for breaking change indicator in type (!)
        is_breaking_from_type = bool(groups.get("breaking"))

        # Parse body and footer
        body_parts = self._parse_body_and_footer(body) if body else (None, None)
        parsed_body, footer = body_parts

        # Check for breaking changes in footer
        breaking_change_desc = self._extract_breaking_change(footer or "")
        is_breaking_from_footer = bool(breaking_change_desc)

        is_breaking = is_breaking_from_type or is_breaking_from_footer

        return ParsedCommit(
            type=commit_type,
            scope=scope,
            description=description,
            body=parsed_body,
            footer=footer,
            is_breaking=is_breaking,
            breaking_change_description=breaking_change_desc,
        )

    def _parse_body_and_footer(self, text: str) -> tuple[Optional[str], Optional[str]]:
        """Parse body and footer from commit message text."""
        if not text:
            return None, None

        # Split on double newlines to separate body from footer
        parts = text.split("\n\n")

        if len(parts) == 1:
            # Only body, no footer
            return parts[0].strip(), None

        # Check if last part looks like footer (contains colons)
        potential_footer = parts[-1]
        if ":" in potential_footer and any(
            line.strip() and ":" in line for line in potential_footer.split("\n")
        ):
            # Last part is footer
            body_parts = parts[:-1]
            body = "\n\n".join(body_parts).strip() if body_parts else None
            footer = potential_footer.strip()
            return body, footer
        else:
            # All parts are body
            return text.strip(), None

    def _extract_breaking_change(self, footer: str) -> Optional[str]:
        """Extract breaking change description from footer."""
        if not footer:
            return None

        for pattern in self.BREAKING_CHANGE_PATTERNS:
            match = pattern.search(footer)
            if match:
                return match.group(1).strip()

        return None

    def is_valid_commit_type(self, commit_type: str) -> bool:
        """Check if commit type is valid."""
        # Common conventional commit types
        valid_types = {
            "feat",
            "fix",
            "docs",
            "style",
            "refactor",
            "perf",
            "test",
            "build",
            "ci",
            "chore",
            "revert",
        }
        return commit_type.lower() in valid_types

    def parse_multiple(self, commit_messages: List[str]) -> List[ParsedCommit]:
        """Parse multiple commit messages."""
        parsed_commits = []
        for message in commit_messages:
            try:
                parsed_commit = self.parse(message)
                parsed_commits.append(parsed_commit)
            except ParseError:
                # Skip invalid commits
                continue
        return parsed_commits



"""Custom exceptions for Git Changelog Maestro."""


class ChangelogError(Exception):
    """Base exception for changelog generation errors."""

    pass


class GitError(ChangelogError):
    """Exception for Git-related errors."""

    pass


class ParseError(ChangelogError):
    """Exception for commit parsing errors."""

    pass


class FormatterError(ChangelogError):
    """Exception for formatting errors."""

    pass


class ConfigError(ChangelogError):
    """Exception for configuration errors."""

    pass



"""Git Changelog Maestro - Generate elegant changelogs from Git commit history."""

__version__ = "0.1.2"
__author__ = "petherldev"
__email__ = "petherl@protonmail.com"

from .core.formatter import ChangelogFormatter
from .core.generator import ChangelogGenerator
from .core.parser import ConventionalCommitParser

__all__ = ["ChangelogGenerator", "ConventionalCommitParser", "ChangelogFormatter"]



"""Git repository operations."""

from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, Iterator, List, Optional

import git
from git import Commit as GitCommit
from git import Repo

from ..utils.exceptions import GitError


@dataclass
class Commit:
    """Represents a Git commit."""

    hash: str
    message: str
    author: str
    email: str
    date: datetime
    is_merge: bool = False

    @classmethod
    def from_git_commit(cls, git_commit: GitCommit) -> "Commit":
        """Create Commit from GitPython commit object."""
        return cls(
            hash=git_commit.hexsha,
            message=(
                git_commit.message.strip().decode()
                if isinstance(git_commit.message, bytes)
                else git_commit.message.strip()
            ),
            author=git_commit.author.name or "Unknown",
            email=git_commit.author.email or "unknown@example.com",
            date=datetime.fromtimestamp(git_commit.committed_date),
            is_merge=len(git_commit.parents) > 1,
        )


@dataclass
class Tag:
    """Represents a Git tag."""

    name: str
    hash: str
    date: datetime
    message: Optional[str] = None

    @classmethod
    def from_git_tag(cls, git_tag: Any) -> "Tag":
        """Create Tag from GitPython tag object."""
        commit = git_tag.commit
        return cls(
            name=git_tag.name,
            hash=commit.hexsha,
            date=datetime.fromtimestamp(commit.committed_date),
            message=(
                git_tag.tag.message if hasattr(git_tag, "tag") and git_tag.tag else None
            ),
        )


class GitRepository:
    """Git repository operations."""

    def __init__(self, repo_path: Path) -> None:
        """Initialize Git repository."""
        self.repo_path = repo_path
        try:
            self.repo = Repo(repo_path)
        except git.InvalidGitRepositoryError:
            raise GitError(f"Not a valid Git repository: {repo_path}")
        except git.NoSuchPathError:
            raise GitError(f"Repository path does not exist: {repo_path}")

    def get_commits(
        self,
        since: Optional[str] = None,
        until: Optional[str] = None,
        include_merges: bool = False,
    ) -> List[Commit]:
        """Get commits from the repository."""
        try:
            # Build revision range
            rev_range = self._build_revision_range(since, until)

            # Get commits
            git_commits = list(self.repo.iter_commits(rev_range))
            commits = [Commit.from_git_commit(gc) for gc in git_commits]

            # Filter merge commits if requested
            if not include_merges:
                commits = [c for c in commits if not c.is_merge]

            return commits

        except git.GitCommandError as e:
            raise GitError(f"Failed to get commits: {e}")

    def get_commits_in_range(
        self,
        since: Optional[str] = None,
        until: Optional[str] = None,
        include_merges: bool = False,
    ) -> List[Commit]:
        """Get commits in a specific tag range."""
        try:
            # Build the revision range for git log
            if since and until:
                # Between two tags
                rev_range = f"{since}..{until}"
            elif since:
                # From tag to HEAD
                rev_range = f"{since}..HEAD"
            elif until:
                # From beginning to tag
                rev_range = until
            else:
                # All commits
                rev_range = "HEAD"

            # Get commits using the range
            git_commits = list(self.repo.iter_commits(rev_range))
            commits = [Commit.from_git_commit(gc) for gc in git_commits]

            # Filter merge commits if requested
            if not include_merges:
                commits = [c for c in commits if not c.is_merge]

            return commits

        except git.GitCommandError as e:
            raise GitError(f"Failed to get commits in range: {e}")

    def get_tags(self) -> List[Tag]:
        """Get all tags from the repository."""
        try:
            git_tags = self.repo.tags
            return [Tag.from_git_tag(gt) for gt in git_tags]
        except git.GitCommandError as e:
            raise GitError(f"Failed to get tags: {e}")

    def get_latest_tag(self) -> Optional[Tag]:
        """Get the latest tag by commit date (not tag creation date)."""
        tags = self.get_tags()
        if not tags:
            return None

        # Sort by commit date (when the commit was made, not when tag was created)
        # This is more reliable than tag creation time
        return max(tags, key=lambda t: t.date)

    def tag_exists(self, tag_name: str) -> bool:
        """Check if a tag exists."""
        try:
            # Check if tag exists in the repository
            for tag in self.repo.tags:
                if tag.name == tag_name:
                    return True
            return False
        except Exception:
            return False

    def get_commits_between_tags(
        self,
        from_tag: Optional[str] = None,
        to_tag: Optional[str] = None,
        include_merges: bool = False,
    ) -> List[Commit]:
        """Get commits between two tags."""
        return self.get_commits(
            since=from_tag,
            until=to_tag,
            include_merges=include_merges,
        )

    def _build_revision_range(
        self,
        since: Optional[str] = None,
        until: Optional[str] = None,
    ) -> str:
        """Build revision range for git log."""
        if since and until:
            return f"{since}..{until}"
        elif since:
            return f"{since}..HEAD"
        elif until:
            return until
        else:
            return "HEAD"

    @property
    def current_branch(self) -> str:
        """Get current branch name."""
        try:
            return self.repo.active_branch.name
        except TypeError:
            # Detached HEAD
            return "HEAD"

    @property
    def remote_url(self) -> Optional[str]:
        """Get remote origin URL."""
        try:
            return self.repo.remotes.origin.url
        except AttributeError:
            return None

    def close(self) -> None:
        """Close the repository to release resources."""
        try:
            if hasattr(self.repo, "close"):
                self.repo.close()
        except Exception:
            pass



"""Command-line interface for Git Changelog Maestro."""

import sys
from pathlib import Path
from typing import List, Optional

import click
from rich.console import Console
from rich.panel import Panel
from rich.text import Text

from .core.config import Config
from .core.generator import ChangelogGenerator
from .utils.exceptions import ChangelogError

console = Console()


def print_error(message: str) -> None:
    """Print error message with rich formatting."""
    console.print(f"[red]Error:[/red] {message}")


def print_success(message: str) -> None:
    """Print success message with rich formatting."""
    console.print(f"[green]✓[/green] {message}")


def print_info(message: str) -> None:
    """Print info message with rich formatting."""
    console.print(f"[blue]ℹ[/blue] {message}")


@click.group(invoke_without_command=True)
@click.option(
    "--repo-path",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, path_type=Path),
    default=Path.cwd(),
    help="Git repository path [default: current directory]",
)
@click.option(
    "--output",
    type=click.Path(path_type=Path),
    default=Path("CHANGELOG.md"),
    help="Output file [default: CHANGELOG.md]",
)
@click.option(
    "--template",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, path_type=Path),
    help="Custom template file",
)
@click.option("--since", help="Generate from specific tag")
@click.option("--until", help="Generate until specific tag")
@click.option("--version-prefix", default="v", help="Version prefix [default: v]")
@click.option(
    "--style",
    type=click.Choice(["markdown", "json", "yaml"]),
    default="markdown",
    help="Output style [default: markdown]",
)
@click.option(
    "--sections",
    multiple=True,
    help="Custom sections to include (can be used multiple times)",
)
@click.option(
    "--exclude",
    multiple=True,
    help="Patterns to exclude (can be used multiple times)",
)
@click.option("--no-merges", is_flag=True, help="Exclude merge commits")
@click.option("--verbose", "-v", is_flag=True, help="Increase output verbosity")
@click.pass_context
def cli(
    ctx: click.Context,
    repo_path: Path,
    output: Path,
    template: Optional[Path],
    since: Optional[str],
    until: Optional[str],
    version_prefix: str,
    style: str,
    sections: List[str],
    exclude: List[str],
    no_merges: bool,
    verbose: bool,
) -> None:
    """Generate elegant changelogs from Git commit history using Conventional Commits."""
    if ctx.invoked_subcommand is None:
        # Default behavior: generate changelog
        ctx.invoke(
            generate,
            repo_path=repo_path,
            output=output,
            template=template,
            since=since,
            until=until,
            version_prefix=version_prefix,
            style=style,
            sections=sections,
            exclude=exclude,
            no_merges=no_merges,
            verbose=verbose,
        )


@cli.command()
@click.option(
    "--repo-path",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, path_type=Path),
    default=Path.cwd(),
    help="Git repository path [default: current directory]",
)
@click.option(
    "--output",
    type=click.Path(path_type=Path),
    default=Path("CHANGELOG.md"),
    help="Output file [default: CHANGELOG.md]",
)
@click.option(
    "--template",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, path_type=Path),
    help="Custom template file",
)
@click.option("--since", help="Generate from specific tag")
@click.option("--until", help="Generate until specific tag")
@click.option("--version-prefix", default="v", help="Version prefix [default: v]")
@click.option(
    "--style",
    type=click.Choice(["markdown", "json", "yaml"]),
    default="markdown",
    help="Output style [default: markdown]",
)
@click.option(
    "--sections",
    multiple=True,
    help="Custom sections to include (can be used multiple times)",
)
@click.option(
    "--exclude",
    multiple=True,
    help="Patterns to exclude (can be used multiple times)",
)
@click.option("--no-merges", is_flag=True, help="Exclude merge commits")
@click.option("--verbose", "-v", is_flag=True, help="Increase output verbosity")
def generate(
    repo_path: Path,
    output: Path,
    template: Optional[Path],
    since: Optional[str],
    until: Optional[str],
    version_prefix: str,
    style: str,
    sections: List[str],
    exclude: List[str],
    no_merges: bool,
    verbose: bool,
) -> None:
    """Generate changelog from Git commit history."""
    try:
        if verbose:
            print_info(f"Repository path: {repo_path}")
            print_info(f"Output file: {output}")
            print_info(f"Style: {style}")

        # Load configuration
        config = Config.load_from_file(repo_path / "pyproject.toml")

        # Override config with CLI options
        if template:
            config.template_path = template
        if since:
            config.since_tag = since
        if until:
            config.until_tag = until
        if version_prefix:
            config.version_prefix = version_prefix
        if style:
            config.output_style = style
        if sections:
            config.sections = list(sections)
        if exclude:
            config.exclude_patterns = list(exclude)
        if no_merges:
            config.include_merges = False

        # Generate changelog
        generator = ChangelogGenerator(repo_path, config)
        changelog_content = generator.generate()

        # Write output
        output.write_text(changelog_content, encoding="utf-8")

        print_success(f"Changelog generated successfully: {output}")

        if verbose:
            console.print(
                Panel(
                    Text(
                        changelog_content[:500] + "..."
                        if len(changelog_content) > 500
                        else changelog_content
                    ),
                    title="Generated Changelog Preview",
                    border_style="green",
                )
            )

    except ChangelogError as e:
        print_error(str(e))
        sys.exit(1)
    except Exception as e:
        print_error(f"Unexpected error: {e}")
        if verbose:
            console.print_exception()
        sys.exit(1)


@cli.command()
@click.option(
    "--repo-path",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, path_type=Path),
    default=Path.cwd(),
    help="Git repository path [default: current directory]",
)
def validate(repo_path: Path) -> None:
    """Validate commit messages against Conventional Commits specification."""
    try:
        from .core.git import GitRepository
        from .core.parser import ConventionalCommitParser

        git_repo = GitRepository(repo_path)
        parser = ConventionalCommitParser()

        commits = git_repo.get_commits()
        valid_count = 0
        invalid_commits = []

        for commit in commits:
            try:
                parser.parse(commit.message)
                valid_count += 1
            except Exception:
                invalid_commits.append(commit)

        total_commits = len(commits)

        console.print(f"\n[bold]Commit Validation Results[/bold]")
        console.print(f"Total commits: {total_commits}")
        console.print(f"[green]Valid commits: {valid_count}[/green]")
        console.print(f"[red]Invalid commits: {len(invalid_commits)}[/red]")

        if invalid_commits:
            console.print("\n[bold red]Invalid commits:[/bold red]")
            for commit in invalid_commits[:10]:  # Show first 10
                console.print(f"  {commit.hash[:8]}: {commit.message[:60]}...")

            if len(invalid_commits) > 10:
                console.print(f"  ... and {len(invalid_commits) - 10} more")

        if len(invalid_commits) > 0:
            sys.exit(1)
        else:
            print_success("All commits follow Conventional Commits specification!")

    except Exception as e:
        print_error(f"Validation failed: {e}")
        sys.exit(1)


@cli.command()
def init() -> None:
    """Initialize changelog configuration in current directory."""
    try:
        config_path = Path("pyproject.toml")

        if config_path.exists():
            # Read existing config and add changelog section
            content = config_path.read_text()
            if "[tool.changelog-maestro]" not in content:
                content += "\n\n[tool.changelog-maestro]\n"
                content += "# Changelog configuration\n"
                content += 'output_file = "CHANGELOG.md"\n'
                content += 'version_prefix = "v"\n'
                content += "include_merges = false\n"
                config_path.write_text(content)
                print_success(
                    "Added changelog configuration to existing pyproject.toml"
                )
            else:
                print_info("Changelog configuration already exists in pyproject.toml")
        else:
            # Create new pyproject.toml with changelog config
            content = """[tool.changelog-maestro]
# Changelog configuration
output_file = "CHANGELOG.md"
version_prefix = "v"
include_merges = false

# Commit types to include in changelog
[tool.changelog-maestro.commit_types]
feat = "Features"
fix = "Bug Fixes"
docs = "Documentation"
style = "Styles"
refactor = "Code Refactoring"
perf = "Performance Improvements"
test = "Tests"
build = "Builds"
ci = "Continuous Integration"
chore = "Chores"
revert = "Reverts"
"""
            config_path.write_text(content)
            print_success("Created pyproject.toml with changelog configuration")

    except Exception as e:
        print_error(f"Failed to initialize configuration: {e}")
        sys.exit(1)


def main() -> None:
    """Main entry point for the CLI."""
    cli()


if __name__ == "__main__":
    main()



"""Tests for changelog formatting."""

from datetime import datetime
from pathlib import Path

import pytest

from changelog_maestro.core.formatter import ChangelogEntry, ChangelogFormatter
from changelog_maestro.core.parser import ParsedCommit
from changelog_maestro.utils.exceptions import FormatterError


class TestChangelogEntry:
    """Test cases for ChangelogEntry."""

    def test_create_entry(self):
        """Test creating a changelog entry."""
        entry = ChangelogEntry("1.0.0", datetime(2023, 1, 1))

        assert entry.version == "1.0.0"
        assert entry.date == datetime(2023, 1, 1)
        assert entry.sections == {}
        assert entry.breaking_changes == []
        assert not entry.has_changes()

    def test_add_commit(self):
        """Test adding commits to entry."""
        entry = ChangelogEntry("1.0.0")

        commit = ParsedCommit(
            type="feat",
            scope=None,
            description="add new feature",
            body=None,
            footer=None,
            is_breaking=False,
        )

        entry.add_commit(commit, "Features")

        assert "Features" in entry.sections
        assert len(entry.sections["Features"]) == 1
        assert entry.sections["Features"][0] == commit
        assert entry.has_changes()

    def test_add_breaking_change(self):
        """Test adding breaking change commit."""
        entry = ChangelogEntry("1.0.0")

        commit = ParsedCommit(
            type="feat",
            scope=None,
            description="breaking change",
            body=None,
            footer=None,
            is_breaking=True,
            breaking_change_description="This breaks the API",
        )

        entry.add_commit(commit, "Features")

        assert len(entry.breaking_changes) == 1
        assert entry.breaking_changes[0] == commit

    def test_to_dict(self):
        """Test converting entry to dictionary."""
        entry = ChangelogEntry("1.0.0", datetime(2023, 1, 1))

        commit = ParsedCommit(
            type="feat",
            scope="api",
            description="add endpoint",
            body="Detailed description",
            footer=None,
            is_breaking=False,
        )

        entry.add_commit(commit, "Features")

        result = entry.to_dict()

        assert result["version"] == "1.0.0"
        assert result["date"] == "2023-01-01T00:00:00"
        assert "Features" in result["sections"]
        assert len(result["sections"]["Features"]) == 1

        commit_dict = result["sections"]["Features"][0]
        assert commit_dict["type"] == "feat"
        assert commit_dict["scope"] == "api"
        assert commit_dict["description"] == "add endpoint"


class TestChangelogFormatter:
    """Test cases for ChangelogFormatter."""

    def setup_method(self):
        """Set up test fixtures."""
        self.formatter = ChangelogFormatter()

    def test_format_markdown(self):
        """Test formatting as Markdown."""
        entry = ChangelogEntry("1.0.0", datetime(2023, 1, 1))

        commit = ParsedCommit(
            type="feat",
            scope=None,
            description="add new feature",
            body=None,
            footer=None,
            is_breaking=False,
        )

        entry.add_commit(commit, "Features")

        result = self.formatter.format_markdown([entry])

        assert "# Changelog" in result
        assert "## [1.0.0] - 2023-01-01" in result
        assert "### Features" in result
        assert "add new feature" in result

    def test_format_json(self):
        """Test formatting as JSON."""
        entry = ChangelogEntry("1.0.0", datetime(2023, 1, 1))

        commit = ParsedCommit(
            type="feat",
            scope=None,
            description="add new feature",
            body=None,
            footer=None,
            is_breaking=False,
        )

        entry.add_commit(commit, "Features")

        result = self.formatter.format_json([entry])

        assert '"changelog"' in result
        assert '"version": "1.0.0"' in result
        assert '"type": "feat"' in result
        assert '"description": "add new feature"' in result

    def test_format_yaml(self):
        """Test formatting as YAML."""
        entry = ChangelogEntry("1.0.0", datetime(2023, 1, 1))

        commit = ParsedCommit(
            type="feat",
            scope=None,
            description="add new feature",
            body=None,
            footer=None,
            is_breaking=False,
        )

        entry.add_commit(commit, "Features")

        result = self.formatter.format_yaml([entry])

        assert "changelog:" in result
        assert "version: 1.0.0" in result
        assert "type: feat" in result
        assert "description: add new feature" in result

    def test_format_with_breaking_changes(self):
        """Test formatting with breaking changes."""
        entry = ChangelogEntry("2.0.0", datetime(2023, 2, 1))

        commit = ParsedCommit(
            type="feat",
            scope="api",
            description="breaking change",
            body=None,
            footer=None,
            is_breaking=True,
            breaking_change_description="This breaks the API",
        )

        entry.add_commit(commit, "Features")

        result = self.formatter.format_markdown([entry])

        assert "⚠ BREAKING CHANGES" in result
        assert "This breaks the API" in result

    def test_format_unsupported_format(self):
        """Test formatting with unsupported format."""
        entry = ChangelogEntry("1.0.0")

        with pytest.raises(FormatterError):
            self.formatter.format([entry], "unsupported")

    def test_format_date_filter(self):
        """Test date formatting filter."""
        date = datetime(2023, 12, 25, 15, 30, 45)
        result = ChangelogFormatter._format_date(date)
        assert result == "2023-12-25"

    def test_format_scope_filter(self):
        """Test scope formatting filter."""
        assert ChangelogFormatter._format_scope("api") == "(api)"
        assert ChangelogFormatter._format_scope(None) == ""
        assert ChangelogFormatter._format_scope("") == ""



"""Configuration management for Git Changelog Maestro."""

import sys
from pathlib import Path
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field

if sys.version_info >= (3, 11):
    import tomllib
else:
    try:
        import tomli as tomllib
    except ImportError:
        import tomllib


class Config(BaseModel):
    """Configuration for changelog generation."""

    output_file: str = Field(default="CHANGELOG.md", description="Output file path")
    template_path: Optional[Path] = Field(
        default=None, description="Custom template file"
    )
    version_prefix: str = Field(default="v", description="Version prefix")
    output_style: str = Field(default="markdown", description="Output format")
    since_tag: Optional[str] = Field(
        default=None, description="Generate from specific tag"
    )
    until_tag: Optional[str] = Field(
        default=None, description="Generate until specific tag"
    )
    sections: List[str] = Field(
        default_factory=list, description="Custom sections to include"
    )
    exclude_patterns: List[str] = Field(
        default_factory=list, description="Patterns to exclude"
    )
    include_merges: bool = Field(default=False, description="Include merge commits")

    # Commit type mappings
    commit_types: Dict[str, str] = Field(
        default_factory=lambda: {
            "feat": "Features",
            "fix": "Bug Fixes",
            "docs": "Documentation",
            "style": "Styles",
            "refactor": "Code Refactoring",
            "perf": "Performance Improvements",
            "test": "Tests",
            "build": "Builds",
            "ci": "Continuous Integration",
            "chore": "Chores",
            "revert": "Reverts",
        },
        description="Mapping of commit types to section names",
    )

    # Breaking change indicators
    breaking_change_indicators: List[str] = Field(
        default_factory=lambda: ["BREAKING CHANGE", "BREAKING-CHANGE"],
        description="Indicators for breaking changes",
    )

    @classmethod
    def load_from_file(cls, config_path: Path) -> "Config":
        """Load configuration from pyproject.toml file."""
        if not config_path.exists():
            return cls()

        try:
            with open(config_path, "rb") as f:
                data = tomllib.load(f)

            changelog_config = data.get("tool", {}).get("changelog-maestro", {})
            return cls(**changelog_config)

        except Exception:
            # If config loading fails, return default config
            return cls()

    def get_section_name(self, commit_type: str) -> str:
        """Get section name for a commit type."""
        return self.commit_types.get(commit_type, commit_type.title())

    def should_include_commit_type(self, commit_type: str) -> bool:
        """Check if a commit type should be included in the changelog."""
        if self.sections:
            return commit_type in self.sections
        return commit_type in self.commit_types



"""Changelog formatting utilities."""

import json
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional

from jinja2 import Environment, FileSystemLoader, Template

from ..utils.exceptions import FormatterError
from .git import Commit, Tag
from .parser import ParsedCommit


class ChangelogEntry:
    """Represents a changelog entry for a version."""

    def __init__(self, version: str, date: Optional[datetime] = None) -> None:
        self.version = version
        self.date = date or datetime.now()
        self.sections: Dict[str, List[ParsedCommit]] = {}
        self.breaking_changes: List[ParsedCommit] = []

    def add_commit(self, commit: ParsedCommit, section_name: str) -> None:
        """Add a commit to the appropriate section."""
        if section_name not in self.sections:
            self.sections[section_name] = []

        self.sections[section_name].append(commit)

        if commit.is_breaking:
            self.breaking_changes.append(commit)

    def has_changes(self) -> bool:
        """Check if this entry has any changes."""
        return bool(self.sections) or bool(self.breaking_changes)

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary for serialization."""
        return {
            "version": self.version,
            "date": self.date.isoformat(),
            "sections": {
                section: [
                    {
                        "type": commit.type,
                        "scope": commit.scope,
                        "description": commit.description,
                        "body": commit.body,
                        "is_breaking": commit.is_breaking,
                        "breaking_change_description": commit.breaking_change_description,
                    }
                    for commit in commits
                ]
                for section, commits in self.sections.items()
            },
            "breaking_changes": [
                {
                    "type": commit.type,
                    "scope": commit.scope,
                    "description": commit.description,
                    "breaking_change_description": commit.breaking_change_description,
                }
                for commit in self.breaking_changes
            ],
        }


class ChangelogFormatter:
    """Formats changelog entries into various output formats."""

    def __init__(self, template_path: Optional[Path] = None) -> None:
        """Initialize formatter with optional custom template."""
        self.template_path = template_path
        self._setup_jinja_env()

    def _setup_jinja_env(self) -> None:
        """Setup Jinja2 environment."""
        if self.template_path and self.template_path.exists():
            # Use custom template
            template_dir = self.template_path.parent
            template_name = self.template_path.name
            self.jinja_env = Environment(loader=FileSystemLoader(template_dir))
            self.template_name = template_name
        else:
            # Use built-in templates
            templates_dir = Path(__file__).parent.parent / "templates"
            self.jinja_env = Environment(loader=FileSystemLoader(templates_dir))
            self.template_name = "default.md.j2"

        # Add custom filters
        self.jinja_env.filters["format_date"] = self._format_date
        self.jinja_env.filters["format_scope"] = self._format_scope

    def format_markdown(self, entries: List[ChangelogEntry]) -> str:
        """Format changelog entries as Markdown."""
        try:
            template = self.jinja_env.get_template(self.template_name)
            return template.render(entries=entries, now=datetime.now())
        except Exception as e:
            raise FormatterError(f"Failed to format Markdown: {e}")

    def format_json(self, entries: List[ChangelogEntry]) -> str:
        """Format changelog entries as JSON."""
        try:
            data = {
                "changelog": [entry.to_dict() for entry in entries],
                "generated_at": datetime.now().isoformat(),
            }
            return json.dumps(data, indent=2, ensure_ascii=False)
        except Exception as e:
            raise FormatterError(f"Failed to format JSON: {e}")

    def format_yaml(self, entries: List[ChangelogEntry]) -> str:
        """Format changelog entries as YAML."""
        try:
            import yaml
        except ImportError:
            raise FormatterError("PyYAML is required for YAML output format")

        try:
            data = {
                "changelog": [entry.to_dict() for entry in entries],
                "generated_at": datetime.now().isoformat(),
            }
            result = yaml.dump(data, default_flow_style=False, allow_unicode=True)
            assert isinstance(result, str)
            return result
        except Exception as e:
            raise FormatterError(f"Failed to format YAML: {e}")

    def format(self, entries: List[ChangelogEntry], output_format: str) -> str:
        """Format changelog entries in the specified format."""
        formatters = {
            "markdown": self.format_markdown,
            "json": self.format_json,
            "yaml": self.format_yaml,
        }

        formatter = formatters.get(output_format.lower())
        if not formatter:
            raise FormatterError(f"Unsupported output format: {output_format}")

        return formatter(entries)

    @staticmethod
    def _format_date(date: datetime) -> str:
        """Format date for display."""
        return date.strftime("%Y-%m-%d")

    @staticmethod
    def _format_scope(scope: Optional[str]) -> str:
        """Format scope for display."""
        return f"({scope})" if scope else ""



"""Tests for the conventional commits parser."""

import pytest

from changelog_maestro.core.parser import ConventionalCommitParser, ParsedCommit
from changelog_maestro.utils.exceptions import ParseError


class TestConventionalCommitParser:
    """Test cases for ConventionalCommitParser."""

    def setup_method(self):
        """Set up test fixtures."""
        self.parser = ConventionalCommitParser()

    def test_parse_simple_commit(self):
        """Test parsing a simple conventional commit."""
        message = "feat: add new feature"
        result = self.parser.parse(message)

        assert result.type == "feat"
        assert result.scope is None
        assert result.description == "add new feature"
        assert result.body is None
        assert result.footer is None
        assert not result.is_breaking

    def test_parse_commit_with_scope(self):
        """Test parsing commit with scope."""
        message = "fix(parser): resolve parsing issue"
        result = self.parser.parse(message)

        assert result.type == "fix"
        assert result.scope == "parser"
        assert result.description == "resolve parsing issue"
        assert not result.is_breaking

    def test_parse_breaking_change_with_exclamation(self):
        """Test parsing breaking change with exclamation mark."""
        message = "feat!: breaking change in API"
        result = self.parser.parse(message)

        assert result.type == "feat"
        assert result.scope is None
        assert result.description == "breaking change in API"
        assert result.is_breaking

    def test_parse_breaking_change_with_scope(self):
        """Test parsing breaking change with scope and exclamation."""
        message = "feat(api)!: breaking change in API"
        result = self.parser.parse(message)

        assert result.type == "feat"
        assert result.scope == "api"
        assert result.description == "breaking change in API"
        assert result.is_breaking

    def test_parse_commit_with_body(self):
        """Test parsing commit with body."""
        message = "feat: add new feature\n\nThis is a detailed description of the feature.\nIt spans multiple lines."
        result = self.parser.parse(message)

        assert result.type == "feat"
        assert result.description == "add new feature"
        assert "detailed description" in result.body
        assert not result.is_breaking

    def test_parse_commit_with_breaking_change_footer(self):
        """Test parsing commit with breaking change in footer."""
        message = "feat: add new feature\n\nThis adds a new feature to the system.\n\nBREAKING CHANGE: This changes the API interface."
        result = self.parser.parse(message)

        assert result.type == "feat"
        assert result.description == "add new feature"
        assert result.is_breaking
        assert "changes the API interface" in result.breaking_change_description

    def test_parse_invalid_commit_format(self):
        """Test parsing invalid commit format."""
        message = "This is not a conventional commit"

        with pytest.raises(ParseError):
            self.parser.parse(message)

    def test_parse_empty_commit(self):
        """Test parsing empty commit message."""
        message = ""

        with pytest.raises(ParseError):
            self.parser.parse(message)

    def test_parse_multiple_commits(self):
        """Test parsing multiple commit messages."""
        messages = [
            "feat: add feature A",
            "fix: resolve bug B",
            "invalid commit message",
            "docs: update documentation",
        ]

        results = self.parser.parse_multiple(messages)

        # Should skip invalid commit
        assert len(results) == 3
        assert results[0].type == "feat"
        assert results[1].type == "fix"
        assert results[2].type == "docs"

    def test_is_valid_commit_type(self):
        """Test commit type validation."""
        assert self.parser.is_valid_commit_type("feat")
        assert self.parser.is_valid_commit_type("fix")
        assert self.parser.is_valid_commit_type("docs")
        assert not self.parser.is_valid_commit_type("invalid")

    def test_parsed_commit_properties(self):
        """Test ParsedCommit properties."""
        commit = ParsedCommit(
            type="feat",
            scope="api",
            description="add new endpoint",
            body=None,
            footer=None,
            is_breaking=False,
        )

        assert commit.is_feature
        assert not commit.is_fix
        assert commit.formatted_scope == "(api)"
        assert commit.short_description == "add new endpoint"

    def test_parsed_commit_without_scope(self):
        """Test ParsedCommit without scope."""
        commit = ParsedCommit(
            type="fix",
            scope=None,
            description="resolve issue",
            body=None,
            footer=None,
            is_breaking=False,
        )

        assert not commit.is_feature
        assert commit.is_fix
        assert commit.formatted_scope == ""



"""Tests for CLI interface."""

from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
from click.testing import CliRunner

from changelog_maestro.cli import cli, generate, init, validate


class TestCLI:
    """Test cases for CLI interface."""

    def setup_method(self):
        """Set up test fixtures."""
        self.runner = CliRunner()

    def test_cli_help(self):
        """Test CLI help output."""
        result = self.runner.invoke(cli, ["--help"])

        assert result.exit_code == 0
        assert "Generate elegant changelogs" in result.output
        assert "--repo-path" in result.output
        assert "--output" in result.output

    @patch("changelog_maestro.cli.ChangelogGenerator")
    def test_generate_command(self, mock_generator_class):
        """Test generate command."""
        # Mock the generator
        mock_generator = MagicMock()
        mock_generator.generate.return_value = "# Changelog\n\n## [1.0.0]\n"
        mock_generator_class.return_value = mock_generator

        with self.runner.isolated_filesystem():
            # Create a dummy git repo directory
            Path("test_repo").mkdir()

            result = self.runner.invoke(
                generate,
                ["--repo-path", "test_repo", "--output", "CHANGELOG.md", "--verbose"],
            )

            assert result.exit_code == 0
            assert "Changelog generated successfully" in result.output

            # Check that changelog file was created
            assert Path("CHANGELOG.md").exists()

    def test_generate_command_error(self):
        """Test generate command with error."""
        with self.runner.isolated_filesystem():
            # Try to generate from non-existent repo
            result = self.runner.invoke(
                generate,
                [
                    "--repo-path",
                    "nonexistent",
                ],
            )

            # Should exit with error code 1, but Click might return 2 for usage errors
            assert result.exit_code in [1, 2]
            assert "Error:" in result.output or "Usage:" in result.output

    @patch("changelog_maestro.core.git.GitRepository")
    @patch("changelog_maestro.core.parser.ConventionalCommitParser")
    def test_validate_command(self, mock_parser_class, mock_git_class):
        """Test validate command."""
        # Mock git repository and parser
        mock_commit = MagicMock()
        mock_commit.message = "feat: add feature"
        mock_commit.hash = "abc123"

        mock_git = MagicMock()
        mock_git.get_commits.return_value = [mock_commit]
        mock_git_class.return_value = mock_git

        mock_parser = MagicMock()
        mock_parser.parse.return_value = MagicMock()
        mock_parser_class.return_value = mock_parser

        with self.runner.isolated_filesystem():
            Path("test_repo").mkdir()

            result = self.runner.invoke(validate, ["--repo-path", "test_repo"])

            assert result.exit_code == 0
            assert "All commits follow Conventional Commits" in result.output

    @patch("changelog_maestro.core.git.GitRepository")
    @patch("changelog_maestro.core.parser.ConventionalCommitParser")
    def test_validate_command_with_invalid_commits(
        self, mock_parser_class, mock_git_class
    ):
        """Test validate command with invalid commits."""
        # Mock git repository
        mock_commit = MagicMock()
        mock_commit.message = "not conventional"
        mock_commit.hash = "abc123"

        mock_git = MagicMock()
        mock_git.get_commits.return_value = [mock_commit]
        mock_git_class.return_value = mock_git

        # Mock parser to raise exception for invalid commit
        mock_parser = MagicMock()
        mock_parser.parse.side_effect = Exception("Invalid format")
        mock_parser_class.return_value = mock_parser

        with self.runner.isolated_filesystem():
            Path("test_repo").mkdir()

            result = self.runner.invoke(validate, ["--repo-path", "test_repo"])

            assert result.exit_code == 1
            assert "Invalid commits:" in result.output

    def test_init_command_new_file(self):
        """Test init command creating new pyproject.toml."""
        with self.runner.isolated_filesystem():
            result = self.runner.invoke(init)

            assert result.exit_code == 0
            assert "Created pyproject.toml" in result.output

            # Check that file was created with correct content
            config_file = Path("pyproject.toml")
            assert config_file.exists()

            content = config_file.read_text()
            assert "[tool.changelog-maestro]" in content
            assert "output_file" in content

    def test_init_command_existing_file(self):
        """Test init command with existing pyproject.toml."""
        with self.runner.isolated_filesystem():
            # Create existing pyproject.toml
            existing_content = """[build-system]
requires = ["setuptools"]
"""
            Path("pyproject.toml").write_text(existing_content)

            result = self.runner.invoke(init)

            assert result.exit_code == 0
            assert "Added changelog configuration" in result.output

            # Check that configuration was added
            content = Path("pyproject.toml").read_text()
            assert "[build-system]" in content  # Original content preserved
            assert "[tool.changelog-maestro]" in content  # New content added

    def test_init_command_existing_config(self):
        """Test init command with existing changelog config."""
        with self.runner.isolated_filesystem():
            # Create pyproject.toml with existing changelog config
            existing_content = """[tool.changelog-maestro]
output_file = "CHANGELOG.md"
"""
            Path("pyproject.toml").write_text(existing_content)

            result = self.runner.invoke(init)

            assert result.exit_code == 0
            assert "already exists" in result.output

    def test_cli_default_behavior(self):
        """Test CLI default behavior (should invoke generate)."""
        with patch("changelog_maestro.cli.ChangelogGenerator") as mock_generator_class:
            mock_generator = MagicMock()
            mock_generator.generate.return_value = "# Changelog\n"
            mock_generator_class.return_value = mock_generator

            with self.runner.isolated_filesystem():
                Path("test_repo").mkdir()

                result = self.runner.invoke(cli, ["--repo-path", "test_repo"])

                assert result.exit_code == 0
                assert "Changelog generated successfully" in result.output



"""Main changelog generator."""

from collections import defaultdict
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional

from ..utils.exceptions import ChangelogError
from .config import Config
from .formatter import ChangelogEntry, ChangelogFormatter
from .git import Commit, GitRepository
from .parser import ConventionalCommitParser, ParsedCommit


class ChangelogGenerator:
    """Main changelog generator class."""

    def __init__(self, repo_path: Path, config: Optional[Config] = None) -> None:
        """Initialize changelog generator."""
        self.repo_path = repo_path
        self.config = config or Config()

        # Initialize components
        self.git_repo = GitRepository(repo_path)
        self.parser = ConventionalCommitParser(
            breaking_change_indicators=self.config.breaking_change_indicators
        )
        self.formatter = ChangelogFormatter(template_path=self.config.template_path)

    def generate(self) -> str:
        """Generate changelog content."""
        try:
            # Get commits from repository using the improved method
            commits = self.git_repo.get_commits(
                since=self.config.since_tag,
                until=self.config.until_tag,
                include_merges=self.config.include_merges,
            )

            if not commits:
                raise ChangelogError("No commits found in the specified range")

            # Parse commits
            parsed_commits = self._parse_commits(commits)

            if not parsed_commits:
                raise ChangelogError("No valid conventional commits found")

            # Group commits by version/tag
            entries = self._group_commits_by_version(parsed_commits)

            # Format changelog
            return self.formatter.format(entries, self.config.output_style)

        except Exception as e:
            if isinstance(e, ChangelogError):
                raise
            raise ChangelogError(f"Failed to generate changelog: {e}")

    def _parse_commits(
        self, commits: List[Commit]
    ) -> List[tuple[Commit, ParsedCommit]]:
        """Parse commits using conventional commits parser."""
        parsed_commits = []

        for commit in commits:
            # Skip commits that match exclude patterns
            if self._should_exclude_commit(commit):
                continue

            try:
                parsed_commit = self.parser.parse(commit.message)

                # Only include commits of specified types
                if self.config.should_include_commit_type(parsed_commit.type):
                    parsed_commits.append((commit, parsed_commit))

            except Exception:
                # Skip commits that don't follow conventional format
                continue

        return parsed_commits

    def _should_exclude_commit(self, commit: Commit) -> bool:
        """Check if commit should be excluded based on patterns."""
        if not self.config.exclude_patterns:
            return False

        message = commit.message.lower()
        for pattern in self.config.exclude_patterns:
            if pattern.lower() in message:
                return True

        return False

    def _group_commits_by_version(
        self, parsed_commits: List[tuple[Commit, ParsedCommit]]
    ) -> List[ChangelogEntry]:
        """Group commits by version/tag."""
        # If we have specific since/until tags, handle them differently
        if self.config.since_tag or self.config.until_tag:
            return self._group_commits_by_tag_range(parsed_commits)
        else:
            return self._group_commits_by_all_tags(parsed_commits)

    def _group_commits_by_tag_range(
        self, parsed_commits: List[tuple[Commit, ParsedCommit]]
    ) -> List[ChangelogEntry]:
        """Group commits when specific tag range is provided."""
        # For tag ranges, we want to show all commits in that range
        # grouped by the tags that exist within that range

        # Get all tags
        all_tags = self.git_repo.get_tags()

        if not all_tags:
            # No tags, create single "Unreleased" entry
            return self._create_unreleased_entry(parsed_commits)

        # Sort tags by date (newest first for display)
        all_tags.sort(key=lambda t: t.date, reverse=True)

        # If we have both since and until, create a single entry for that range
        if self.config.since_tag and self.config.until_tag:
            version_name = f"{self.config.since_tag}..{self.config.until_tag}"
            entry = ChangelogEntry(version_name, datetime.now())

            for commit, parsed_commit in parsed_commits:
                section_name = self.config.get_section_name(parsed_commit.type)
                entry.add_commit(parsed_commit, section_name)

            return [entry] if entry.has_changes() else []

        # Otherwise, group by individual tags in the range
        entries = []

        # Find relevant tags based on the range
        relevant_tags = []
        for tag in all_tags:
            include_tag = True

            # Check if tag is within our range
            if self.config.since_tag:
                since_tag = next(
                    (t for t in all_tags if t.name == self.config.since_tag), None
                )
                if since_tag and tag.date <= since_tag.date:
                    include_tag = False

            if self.config.until_tag:
                until_tag = next(
                    (t for t in all_tags if t.name == self.config.until_tag), None
                )
                if until_tag and tag.date > until_tag.date:
                    include_tag = False

            if include_tag:
                relevant_tags.append(tag)

        # Create entries for relevant tags
        for tag in relevant_tags:
            # For now, include all commits for each tag in range
            # This is a simplified approach - in practice you might want
            # to be more precise about which commits belong to which tag
            if parsed_commits:
                entry = self._create_version_entry(tag.name, tag.date, parsed_commits)
                if entry.has_changes():
                    entries.append(entry)
                    break  # Only create one entry to avoid duplication

        # If no relevant tags found, create unreleased entry
        if not entries and parsed_commits:
            entries = self._create_unreleased_entry(parsed_commits)

        return entries

    def _group_commits_by_all_tags(
        self, parsed_commits: List[tuple[Commit, ParsedCommit]]
    ) -> List[ChangelogEntry]:
        """Group commits by all tags when no specific range is provided."""
        # Get tags from repository
        tags = self.git_repo.get_tags()

        if not tags:
            # No tags, create single "Unreleased" entry
            return self._create_unreleased_entry(parsed_commits)

        # Sort tags by date (oldest first)
        tags.sort(key=lambda t: t.date)

        entries = []
        processed_commits = set()

        # Process each tag (oldest to newest)
        for i, tag in enumerate(tags):
            # Get the date range for this tag
            if i == 0:
                # First tag - all commits up to this tag
                start_date = datetime.min
            else:
                # Subsequent tags - commits after previous tag
                start_date = tags[i - 1].date

            end_date = tag.date

            # Get commits for this version
            version_commits = [
                (commit, parsed)
                for commit, parsed in parsed_commits
                if start_date < commit.date <= end_date
                and (commit.hash, parsed.type) not in processed_commits
            ]

            if version_commits:
                entry = self._create_version_entry(tag.name, tag.date, version_commits)
                entries.append(entry)

                # Mark commits as processed
                for commit, parsed in version_commits:
                    processed_commits.add((commit.hash, parsed.type))

        # Add unreleased commits (commits newer than the latest tag)
        if tags:
            latest_tag_date = max(tag.date for tag in tags)
            unreleased_commits = [
                (commit, parsed)
                for commit, parsed in parsed_commits
                if commit.date > latest_tag_date
                and (commit.hash, parsed.type) not in processed_commits
            ]

            if unreleased_commits:
                unreleased_entry = self._create_unreleased_entry(unreleased_commits)
                entries = unreleased_entry + entries

        # Sort entries by version (newest first for display)
        entries.sort(key=lambda e: e.date, reverse=True)

        return entries

    def _create_version_entry(
        self, version: str, date: datetime, commits: List[tuple[Commit, ParsedCommit]]
    ) -> ChangelogEntry:
        """Create a changelog entry for a specific version."""
        # Remove version prefix if present
        clean_version = version
        if version.startswith(self.config.version_prefix):
            clean_version = version[len(self.config.version_prefix) :]

        entry = ChangelogEntry(clean_version, date)

        # Group commits by type
        for commit, parsed_commit in commits:
            section_name = self.config.get_section_name(parsed_commit.type)
            entry.add_commit(parsed_commit, section_name)

        return entry

    def _create_unreleased_entry(
        self, commits: List[tuple[Commit, ParsedCommit]]
    ) -> List[ChangelogEntry]:
        """Create changelog entry for unreleased commits."""
        if not commits:
            return []

        entry = ChangelogEntry("Unreleased", datetime.now())

        # Group commits by type
        for commit, parsed_commit in commits:
            section_name = self.config.get_section_name(parsed_commit.type)
            entry.add_commit(parsed_commit, section_name)

        return [entry]



"""Tests for Git repository operations."""

import gc
import tempfile
import time
from datetime import datetime
from pathlib import Path

import git
import pytest

from changelog_maestro.core.git import Commit, GitRepository, Tag
from changelog_maestro.utils.exceptions import GitError


class TestGitRepository:
    """Test cases for GitRepository."""

    def test_init_valid_repo(self, temp_repo: Path):
        """Test initializing with valid repository."""
        git_repo = GitRepository(temp_repo)
        assert git_repo.repo_path == temp_repo
        assert isinstance(git_repo.repo, git.Repo)

    def test_init_invalid_repo(self):
        """Test initializing with invalid repository."""
        with tempfile.TemporaryDirectory() as temp_dir:
            invalid_path = Path(temp_dir) / "not_a_repo"
            invalid_path.mkdir()

            with pytest.raises(GitError):
                GitRepository(invalid_path)

    def test_init_nonexistent_path(self):
        """Test initializing with nonexistent path."""
        nonexistent_path = Path("/nonexistent/path")

        with pytest.raises(GitError):
            GitRepository(nonexistent_path)

    def test_get_commits(self, git_repo: GitRepository):
        """Test getting commits from repository."""
        commits = git_repo.get_commits()

        assert len(commits) >= 1  # At least initial commit
        assert all(isinstance(c, Commit) for c in commits)
        assert commits[0].message == "Initial commit"

    def test_get_commits_exclude_merges(self, temp_repo: Path):
        """Test excluding merge commits."""
        repo = git.Repo(temp_repo)
        git_repo = None

        try:
            # Create a branch and merge commit
            feature_branch = repo.create_head("feature")
            feature_branch.checkout()

            # Add commit to feature branch
            test_file = temp_repo / "feature.txt"
            test_file.write_text("feature content")
            repo.index.add([str(test_file)])
            repo.index.commit("feat: add feature")

            # Switch back to main and merge (this should create a merge commit)
            main_branch = repo.heads[0]  # Get the main branch (master/main)
            main_branch.checkout()

            # Force a merge commit
            repo.git.merge("feature", "--no-ff")

            git_repo = GitRepository(temp_repo)

            # Get all commits (including merges)
            all_commits = git_repo.get_commits(include_merges=True)

            # Get commits excluding merges
            no_merge_commits = git_repo.get_commits(include_merges=False)

            # Should have at least one merge commit
            merge_commits = [c for c in all_commits if c.is_merge]
            if merge_commits:
                assert len(all_commits) > len(no_merge_commits)
            else:
                # If no merge commit was created, counts should be equal
                assert len(all_commits) == len(no_merge_commits)
        finally:
            if git_repo:
                git_repo.close()
            repo.close()
            gc.collect()
            time.sleep(0.1)

    def test_get_tags(self, temp_repo: Path):
        """Test getting tags from repository."""
        repo = git.Repo(temp_repo)
        git_repo = None

        try:
            # Create a tag
            repo.create_tag("v1.0.0", message="Release v1.0.0")

            git_repo = GitRepository(temp_repo)
            tags = git_repo.get_tags()

            assert len(tags) == 1
            assert isinstance(tags[0], Tag)
            assert tags[0].name == "v1.0.0"
        finally:
            if git_repo:
                git_repo.close()
            repo.close()
            gc.collect()
            time.sleep(0.1)

    def test_get_latest_tag(self, temp_repo: Path):
        """Test getting latest tag."""
        repo = git.Repo(temp_repo)
        git_repo = None

        try:
            # Get the initial commit
            initial_commit = repo.head.commit

            # Create first tag on initial commit
            tag1 = repo.create_tag("v1.0.0", ref=initial_commit)

            # Wait a moment to ensure different timestamps
            time.sleep(1.1)

            # Add another commit - this will have a later timestamp
            test_file = temp_repo / "test.txt"
            test_file.write_text("test content")
            repo.index.add([str(test_file)])
            new_commit = repo.index.commit("Add test file")

            # Create second tag on the newer commit
            tag2 = repo.create_tag("v1.1.0", ref=new_commit)

            git_repo = GitRepository(temp_repo)

            # Verify the commits have different timestamps
            commit1_date = datetime.fromtimestamp(tag1.commit.committed_date)
            commit2_date = datetime.fromtimestamp(tag2.commit.committed_date)

            # The second commit should be newer
            assert (
                commit2_date > commit1_date
            ), f"Second commit ({commit2_date}) should be newer than first ({commit1_date})"

            latest_tag = git_repo.get_latest_tag()

            assert latest_tag is not None
            # The latest tag should be v1.1.0 since it points to a later commit
            assert (
                latest_tag.name == "v1.1.0"
            ), f"Expected v1.1.0 but got {latest_tag.name}"
        finally:
            if git_repo:
                git_repo.close()
            repo.close()
            gc.collect()
            time.sleep(0.1)

    def test_get_latest_tag_no_tags(self, git_repo: GitRepository):
        """Test getting latest tag when no tags exist."""
        latest_tag = git_repo.get_latest_tag()
        assert latest_tag is None

    def test_tag_exists(self, temp_repo: Path):
        """Test checking if tag exists."""
        repo = git.Repo(temp_repo)
        git_repo = None

        try:
            repo.create_tag("v1.0.0")

            git_repo = GitRepository(temp_repo)

            assert git_repo.tag_exists("v1.0.0")
            assert not git_repo.tag_exists("v2.0.0")
        finally:
            if git_repo:
                git_repo.close()
            repo.close()
            gc.collect()
            time.sleep(0.1)

    def test_current_branch(self, git_repo: GitRepository):
        """Test getting current branch name."""
        branch_name = git_repo.current_branch
        assert branch_name in ["master", "main"]  # Default branch names

    def test_remote_url_no_remote(self, git_repo: GitRepository):
        """Test getting remote URL when no remote exists."""
        remote_url = git_repo.remote_url
        assert remote_url is None


class TestCommit:
    """Test cases for Commit dataclass."""

    def test_from_git_commit(self, temp_repo: Path):
        """Test creating Commit from GitPython commit."""
        repo = git.Repo(temp_repo)

        try:
            git_commit = list(repo.iter_commits())[0]

            commit = Commit.from_git_commit(git_commit)

            assert commit.hash == git_commit.hexsha
            assert commit.message == git_commit.message.strip()
            assert commit.author == git_commit.author.name
            assert commit.email == git_commit.author.email
            assert isinstance(commit.date, datetime)
            assert not commit.is_merge  # Initial commit is not a merge
        finally:
            repo.close()
            gc.collect()
            time.sleep(0.1)


class TestTag:
    """Test cases for Tag dataclass."""

    def test_from_git_tag(self, temp_repo: Path):
        """Test creating Tag from GitPython tag."""
        repo = git.Repo(temp_repo)

        try:
            git_tag = repo.create_tag("v1.0.0", message="Release v1.0.0")

            tag = Tag.from_git_tag(git_tag)

            assert tag.name == "v1.0.0"
            assert tag.hash == git_tag.commit.hexsha
            assert isinstance(tag.date, datetime)
        finally:
            repo.close()
            gc.collect()
            time.sleep(0.1)
